# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: qrlbase.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='qrlbase.proto',
  package='qrl',
  syntax='proto3',
  serialized_pb=_b('\n\rqrlbase.proto\x12\x03qrl\"\x10\n\x0eGetNodeInfoReq\"5\n\x0fGetNodeInfoResp\x12\x0f\n\x07version\x18\x01 \x01(\t\x12\x11\n\tgrpcProto\x18\x02 \x01(\t2B\n\x04\x42\x61se\x12:\n\x0bGetNodeInfo\x12\x13.qrl.GetNodeInfoReq\x1a\x14.qrl.GetNodeInfoResp\"\x00\x62\x06proto3')
)




_GETNODEINFOREQ = _descriptor.Descriptor(
  name='GetNodeInfoReq',
  full_name='qrl.GetNodeInfoReq',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22,
  serialized_end=38,
)


_GETNODEINFORESP = _descriptor.Descriptor(
  name='GetNodeInfoResp',
  full_name='qrl.GetNodeInfoResp',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='version', full_name='qrl.GetNodeInfoResp.version', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='grpcProto', full_name='qrl.GetNodeInfoResp.grpcProto', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=40,
  serialized_end=93,
)

DESCRIPTOR.message_types_by_name['GetNodeInfoReq'] = _GETNODEINFOREQ
DESCRIPTOR.message_types_by_name['GetNodeInfoResp'] = _GETNODEINFORESP
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

GetNodeInfoReq = _reflection.GeneratedProtocolMessageType('GetNodeInfoReq', (_message.Message,), dict(
  DESCRIPTOR = _GETNODEINFOREQ,
  __module__ = 'qrlbase_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetNodeInfoReq)
  ))
_sym_db.RegisterMessage(GetNodeInfoReq)

GetNodeInfoResp = _reflection.GeneratedProtocolMessageType('GetNodeInfoResp', (_message.Message,), dict(
  DESCRIPTOR = _GETNODEINFORESP,
  __module__ = 'qrlbase_pb2'
  # @@protoc_insertion_point(class_scope:qrl.GetNodeInfoResp)
  ))
_sym_db.RegisterMessage(GetNodeInfoResp)



_BASE = _descriptor.ServiceDescriptor(
  name='Base',
  full_name='qrl.Base',
  file=DESCRIPTOR,
  index=0,
  options=None,
  serialized_start=95,
  serialized_end=161,
  methods=[
  _descriptor.MethodDescriptor(
    name='GetNodeInfo',
    full_name='qrl.Base.GetNodeInfo',
    index=0,
    containing_service=None,
    input_type=_GETNODEINFOREQ,
    output_type=_GETNODEINFORESP,
    options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_BASE)

DESCRIPTOR.services_by_name['Base'] = _BASE

try:
  # THESE ELEMENTS WILL BE DEPRECATED.
  # Please use the generated *_pb2_grpc.py files instead.
  import grpc
  from grpc.beta import implementations as beta_implementations
  from grpc.beta import interfaces as beta_interfaces
  from grpc.framework.common import cardinality
  from grpc.framework.interfaces.face import utilities as face_utilities


  class BaseStub(object):
    # missing associated documentation comment in .proto file
    pass

    def __init__(self, channel):
      """Constructor.

      Args:
        channel: A grpc.Channel.
      """
      self.GetNodeInfo = channel.unary_unary(
          '/qrl.Base/GetNodeInfo',
          request_serializer=GetNodeInfoReq.SerializeToString,
          response_deserializer=GetNodeInfoResp.FromString,
          )


  class BaseServicer(object):
    # missing associated documentation comment in .proto file
    pass

    def GetNodeInfo(self, request, context):
      # missing associated documentation comment in .proto file
      context.set_code(grpc.StatusCode.UNIMPLEMENTED)
      context.set_details('Method not implemented!')
      raise NotImplementedError('Method not implemented!')


  def add_BaseServicer_to_server(servicer, server):
    rpc_method_handlers = {
        'GetNodeInfo': grpc.unary_unary_rpc_method_handler(
            servicer.GetNodeInfo,
            request_deserializer=GetNodeInfoReq.FromString,
            response_serializer=GetNodeInfoResp.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        'qrl.Base', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


  class BetaBaseServicer(object):
    """The Beta API is deprecated for 0.15.0 and later.

    It is recommended to use the GA API (classes and functions in this
    file not marked beta) for all further purposes. This class was generated
    only to ease transition from grpcio<0.15.0 to grpcio>=0.15.0."""
    # missing associated documentation comment in .proto file
    def GetNodeInfo(self, request, context):
      # missing associated documentation comment in .proto file
      context.code(beta_interfaces.StatusCode.UNIMPLEMENTED)


  class BetaBaseStub(object):
    """The Beta API is deprecated for 0.15.0 and later.

    It is recommended to use the GA API (classes and functions in this
    file not marked beta) for all further purposes. This class was generated
    only to ease transition from grpcio<0.15.0 to grpcio>=0.15.0."""
    # missing associated documentation comment in .proto file
    def GetNodeInfo(self, request, timeout, metadata=None, with_call=False, protocol_options=None):
      # missing associated documentation comment in .proto file
      raise NotImplementedError()
    GetNodeInfo.future = None


  def beta_create_Base_server(servicer, pool=None, pool_size=None, default_timeout=None, maximum_timeout=None):
    """The Beta API is deprecated for 0.15.0 and later.

    It is recommended to use the GA API (classes and functions in this
    file not marked beta) for all further purposes. This function was
    generated only to ease transition from grpcio<0.15.0 to grpcio>=0.15.0"""
    request_deserializers = {
      ('qrl.Base', 'GetNodeInfo'): GetNodeInfoReq.FromString,
    }
    response_serializers = {
      ('qrl.Base', 'GetNodeInfo'): GetNodeInfoResp.SerializeToString,
    }
    method_implementations = {
      ('qrl.Base', 'GetNodeInfo'): face_utilities.unary_unary_inline(servicer.GetNodeInfo),
    }
    server_options = beta_implementations.server_options(request_deserializers=request_deserializers, response_serializers=response_serializers, thread_pool=pool, thread_pool_size=pool_size, default_timeout=default_timeout, maximum_timeout=maximum_timeout)
    return beta_implementations.server(method_implementations, options=server_options)


  def beta_create_Base_stub(channel, host=None, metadata_transformer=None, pool=None, pool_size=None):
    """The Beta API is deprecated for 0.15.0 and later.

    It is recommended to use the GA API (classes and functions in this
    file not marked beta) for all further purposes. This function was
    generated only to ease transition from grpcio<0.15.0 to grpcio>=0.15.0"""
    request_serializers = {
      ('qrl.Base', 'GetNodeInfo'): GetNodeInfoReq.SerializeToString,
    }
    response_deserializers = {
      ('qrl.Base', 'GetNodeInfo'): GetNodeInfoResp.FromString,
    }
    cardinalities = {
      'GetNodeInfo': cardinality.Cardinality.UNARY_UNARY,
    }
    stub_options = beta_implementations.stub_options(host=host, metadata_transformer=metadata_transformer, request_serializers=request_serializers, response_deserializers=response_deserializers, thread_pool=pool, thread_pool_size=pool_size)
    return beta_implementations.dynamic_stub(channel, 'qrl.Base', cardinalities, options=stub_options)
except ImportError:
  pass
# @@protoc_insertion_point(module_scope)
